import numpy as np
from typing import Optional, Union, List, Set, Tuple, Any, Callable, Type
from math import floor

from spode.core import Circuit

__all__ = ['get_supported_config', 'generate']

# stores all the supported configurations for automatic generation
_config_size_dict = {'square_1': [2, 'A basic square mesh of size N by M. Two integers are needed to define this '
                                     'configuration, representing #Row and #Col.'],
                     'hexagonal_1': [1, 'A concentric hexagonal mesh with N levels. One integer is needed to define '
                                        'this configuration, representing #level.'],
                     'triangular_1': [2, 'A triangular mesh structure of size N by M (close to diamond). Two integers '
                                         'are needed to define this configuration, representing #Row and #Col.']}


def get_supported_config():
    """Print the supported configurations."""

    print("\nAutomatic schematic generation is supported for the following configurations:")
    for k, v in _config_size_dict.items():
        print("### config = '%s', len(size) = %d. %s" % (k, v[0], v[1]))
    print("\n")


def generate(config: str, size: List[int], init_dict: dict, tbu_model: Optional[str] = 'tbum'):
    """Generate the circuit netlist

    Generate the circuit netlist based on self.config, self.size, and self.tbu_model.

    :param config: A string specify the configuration, e.g., 'square_1'.
    :param size: A list of integer specify the size of the programmable photonic circuit.
                 The length of it depends on the variable 'config'.1
    :param init_dict: specify the initial values for all TBUs.
    :param tbu_model: specify which tbu model to use, e.g., 'tbum', 'tbuo', 'tbut'.
    :return: a dict contains all circuit elements.
    """
    return Generator().generate(config, size, init_dict, tbu_model)


class Generator(object):
    """Automatic programmable photonics circuit schematic generator"""

    def __init__(self):
        pass

    def generate(self, config: str, size: List[int], init_dict: dict, tbu_model: Optional[str] = 'tbum') -> dict:
        """Generate the circuit netlist

        Generate the circuit netlist based on self.config, self.size, and self.tbu_model.

        :param config: A string specify the configuration, e.g., 'square_1'.
        :param size: A list of integer specify the size of the programmable photonic circuit.
                     The length of it depends on the variable 'config'.1
        :param init_dict: specify the initial values for all TBUs.
        :param tbu_model: specify which tbu model to use, e.g., 'tbum', 'tbuo', 'tbut'.
        :return: a dict contains all circuit elements.
        """

        self.config = config
        self.size = size
        self.tbu_model = tbu_model

        if self.config not in _config_size_dict.keys():
            raise RuntimeError("The provided configuration is not supported for automatic schematic generation.")
        elif _config_size_dict[self.config][0] != len(self.size):
            raise RuntimeError("%d integer(s) required to define this configuration, but %d is provided."
                               % (_config_size_dict[self.config][0], len(self.size)))
        else:
            return getattr(self, '_generate_' + self.config)(init_dict)

    def _generate_square_1(self, init_dict) -> dict:
        """Generate circuit elements for a N by M square mesh (will be called in generate function).

        :param init_dict: specify the initial values for all TBUs
        :return" a dict contains all circuit elements.

        """
        circuit_element = {}

        # our 'ln' and 'rn' follow:
        #     -------theta
        #     -------phi
        # |  |
        # |  |
        # phi theta

        # process all vertical TBUs
        for i in range(1, self.size[0] + 1):
            for j in range(0, self.size[1] + 1):
                cur_tbu_model = self.tbu_model + '_' + str(i) + '#' + str(j) + '_' + str(i) + '#' + str(j + 1) + '_v'
                node_dict = {
                    'ln': ['n' + '_' + str(i) + '#' + str(j + 1) + '_' + 'tl',
                           'n' + '_' + str(i) + '#' + str(j) + '_' + 'tr'],
                    'rn': ['n' + '_' + str(i) + '#' + str(j + 1) + '_' + 'bl',
                           'n' + '_' + str(i) + '#' + str(j) + '_' + 'br']}
                circuit_element[cur_tbu_model] = {**node_dict, **init_dict}

        # process all horizontal TBUs
        for i in range(0, self.size[0] + 1):
            for j in range(1, self.size[1] + 1):
                cur_tbu_model = self.tbu_model + '_' + str(i) + '#' + str(j) + '_' + str(i + 1) + '#' + str(j) + '_h'
                node_dict = {
                    'ln': ['n' + '_' + str(i) + '#' + str(j) + '_' + 'bl',
                           'n' + '_' + str(i + 1) + '#' + str(j) + '_' + 'tl'],
                    'rn': ['n' + '_' + str(i) + '#' + str(j) + '_' + 'br',
                           'n' + '_' + str(i + 1) + '#' + str(j) + '_' + 'tr']}
                circuit_element[cur_tbu_model] = {**node_dict, **init_dict}

        return circuit_element

    def _generate_hexagonal_1(self, init_dict) -> dict:
        """Generate circuit elements for a concentric hexagonal mesh (will be called in generate function).

        :param init_dict: specify the initial values for all TBUs
        :return" a dict contains all circuit elements.

        """

        def _helper_mod(dividend: int, divisor: int) -> int:
            """A helper modulus function

            For example, if divisor = 4, then:
            it return 1, given dividend = 1
            it return 2, given dividend = 2
            it return 3, given dividend = 3
            it return 4, given dividend = 4
            it return 1, given dividend = 5
            """
            return (dividend - 1) % divisor + 1

        # level = 1,2,3,...
        num_tbu_func = lambda level: max(6 * (level - 1), 6)
        num_radial_edge_func = lambda level: 6 * (level - 1)
        num_normal_edge_func = lambda level: 12 * level - 6

        circuit_element = {}

        magic_connect = [{'ln': ['p1', 'p3'], 'rn': ['p6', 'p4']},
                         {'ln': ['p2', 'p4'], 'rn': ['p1', 'p5']},
                         {'ln': ['p3', 'p5'], 'rn': ['p2', 'p6']},
                         {'ln': ['p4', 'p6'], 'rn': ['p3', 'p1']},
                         {'ln': ['p5', 'p1'], 'rn': ['p4', 'p2']},
                         {'ln': ['p6', 'p2'], 'rn': ['p5', 'p3']}, ]

        magic_index = np.array([4, 5, 0, 1, 2, 3], dtype=np.int)
        index_add = magic_index + 1

        num_layer = self.size[0]

        for i in range(1, num_layer + 1):
            num_tbus = num_tbu_func(i)
            num_radial_edges = num_radial_edge_func(i)

            # deal with radial edges
            for j in range(1, num_radial_edges + 1):
                # map {1,2,...,num_radial_edge} to {2,3,..., 1}
                ind_j = j % num_radial_edges + 1

                cur_tbu_model = self.tbu_model + '_' + str(i) + '#' + str(j) + '_' + str(i) + '#' + str(ind_j) + '_rd'

                indicator = (j - 1) // (i - 1)  # which one of the magic_connect

                ln = ['n' + '_' + str(i) + '#' + str(ind_j) + '_' + magic_connect[indicator]['ln'][0],
                      'n' + '_' + str(i) + '#' + str(j) + '_' + magic_connect[indicator]['ln'][1]]
                rn = ['n' + '_' + str(i) + '#' + str(ind_j) + '_' + magic_connect[indicator]['rn'][0],
                      'n' + '_' + str(i) + '#' + str(j) + '_' + magic_connect[indicator]['rn'][1]]

                node_dict = {'ln': ln, 'rn': rn}
                circuit_element[cur_tbu_model] = {**node_dict, **init_dict}

            # deal with normal edges
            start_ind_list = magic_index * (i - 1) + 1
            end_ind_list = (magic_index + 2) * (i - 1) + 1
            for m in range(len(start_ind_list)):
                for k in range(start_ind_list[m], end_ind_list[m] + 1):
                    ind_k = _helper_mod(k, num_tbus)
                    ind_k_next_level = _helper_mod(k + index_add[m], num_tbu_func(i + 1))
                    cur_tbu_model = self.tbu_model + '_' + str(i) + '#' + str(ind_k) + '_' + str(i + 1) + '#' + str(
                        ind_k_next_level) + '_nd'

                    ln = ['n' + '_' + str(i) + '#' + str(ind_k) + '_' + magic_connect[m]['ln'][0],
                          'n' + '_' + str(i + 1) + '#' + str(ind_k_next_level) + '_' + magic_connect[m]['ln'][1]]
                    rn = ['n' + '_' + str(i) + '#' + str(ind_k) + '_' + magic_connect[m]['rn'][0],
                          'n' + '_' + str(i + 1) + '#' + str(ind_k_next_level) + '_' + magic_connect[m]['rn'][1]]

                    node_dict = {'ln': ln, 'rn': rn}
                    circuit_element[cur_tbu_model] = {**node_dict, **init_dict}

        return circuit_element

    def _generate_triangular_1(self, init_dict) -> dict:
        """Generate circuit elements for a triangular mesh (will be called in generate function).

                :param init_dict: specify the initial values for all TBUs
                :return" a dict contains all circuit elements.

                """
        magic_connect = [{'ln': ['p1', 'p3'], 'rn': ['p3', 'p1']},
                         {'ln': ['p3', 'p2'], 'rn': ['p2', 'p3']},
                         {'ln': ['p2', 'p1'], 'rn': ['p1', 'p2']}]

        circuit_element = {}
        for i in range(1, self.size[0] + 1):
            # deal with those tilted TBUs
            for j in range(0, self.size[1] + 1):
                cur_tbu_model = self.tbu_model + '_' + str(i) + '#' + str(j) + '_' + str(i) + '#' + str(j + 1) + '_t'
                ln = ['n' + '_' + str(i) + '#' + str(j) + '_' + magic_connect[j % 2]['ln'][0],
                      'n' + '_' + str(i) + '#' + str(j + 1) + '_' + magic_connect[j % 2]['ln'][1]]
                rn = ['n' + '_' + str(i) + '#' + str(j) + '_' + magic_connect[j % 2]['rn'][0],
                      'n' + '_' + str(i) + '#' + str(j + 1) + '_' + magic_connect[j % 2]['rn'][1]]
                node_dict = {'ln': ln, 'rn': rn}
                circuit_element[cur_tbu_model] = {**node_dict, **init_dict}

        for i in range(1, self.size[0] + 2):
            # deal with horizontal TBUs
            for j in range(1, floor((self.size[1] - 1) / 2) + 2):
                ind_j = 2 * j - 1
                cur_tbu_model = self.tbu_model + '_' + str(i - 1) + '#' + str(ind_j + 1) + '_' + str(i) + '#' + str(
                    ind_j) + '_h'
                ln = ['n' + '_' + str(i - 1) + '#' + str(ind_j + 1) + '_' + magic_connect[2]['ln'][0],
                      'n' + '_' + str(i) + '#' + str(ind_j) + '_' + magic_connect[2]['ln'][1]]
                rn = ['n' + '_' + str(i - 1) + '#' + str(ind_j + 1) + '_' + magic_connect[2]['rn'][0],
                      'n' + '_' + str(i) + '#' + str(ind_j) + '_' + magic_connect[2]['rn'][1]]
                node_dict = {'ln': ln, 'rn': rn}
                circuit_element[cur_tbu_model] = {**node_dict, **init_dict}
        return circuit_element


if __name__ == '__main__':
    print("hello worlds")
    a = Generator()
    circuit_element = a.generate('triangular_1', [2, 2], {'theta': 0.0 * np.pi, 'phi': 0.0 * np.pi, 'l': 250e-6})
    print(circuit_element)

    print(circuit_element.keys())

    circ_instance = Circuit(circuit_element=circuit_element,
                            mode_info={'neff': 4.0},
                            omega=np.linspace(192.5, 193.5, 3) * 1e12 * 2 * np.pi,
                            srce_node={'n_0#2_p2': 1.0},
                            prob_node=['n_3#1_p2'],
                            deri_node=[],
                            deri_vari=[])

    response = circ_instance.solve(False)
    print(np.abs(response[0, :, :]))



import numpy as np
import matplotlib.pyplot as plt
from spode.core.solver import Circuit
from typing import Optional, Union, List, Set, Tuple, Any, Callable, Type
from spode.util.generator import generate
from scipy.spatial import ConvexHull
import matplotlib.colors as mcolors

__all__ = ['visualize']


def _cell_place_function_list(placement: str) -> Callable:
    """Return a callable function calculating the coordinate of cell center based on the specified placement

    :param placement: a string represents the placement.
    :return: a callable function.
    """

    if 'square_1' == placement.lower():
        def cell_place_function(cell_name, tbu_length, tbu_width):
            x = ([int(ele) for ele in cell_name.split('#')])
            return np.array([tbu_width * (x[1] + 1) + tbu_length * (x[1] + 0.5),
                             tbu_width * (x[0] + 1) + tbu_length * (x[0] + 0.5)])
    elif 'hexagonal_1' == placement.lower():
        def cell_place_function(cell_name, tbu_length, tbu_width):
            layer_index, cell_index = ([int(ele) for ele in cell_name.split('#')])
            if layer_index == 1:
                return np.zeros(2, )
            unit_length = (3 ** 0.5) * tbu_length + tbu_width
            vertices = [1 + ele for ele in np.arange(7) * (layer_index - 1)]
            length = (layer_index - 1) * unit_length
            angles = np.array([np.pi, np.pi * 2 / 3.0, np.pi / 3.0, 0.0, -np.pi / 3.0, -np.pi * 2 / 3.0, -np.pi])
            coordinates = [length * np.array([np.cos(angle), -np.sin(angle)]) for angle in angles]
            for i in range(len(vertices) - 1):
                if vertices[i] <= cell_index < vertices[i + 1]:
                    start, end = coordinates[i], coordinates[i + 1]
                    ratio = (cell_index - vertices[i]) / (layer_index - 1)
                    x = start * (1 - ratio) + end * ratio
                    return x
    elif 'triangular_1' == placement.lower():
        def cell_place_function(cell_name, tbu_length, tbu_width):
            row_index, column_index = ([int(ele) for ele in cell_name.split('#')])
            unit_length = tbu_length / (3 ** 0.5) + tbu_width
            x = 0.5 * (3 ** 0.5) * (row_index - 1) * unit_length + (column_index - 1) * 0.5 * (3 ** 0.5) * unit_length
            y = 1.5 * (row_index - 1) * unit_length + 0.5 * (column_index % 2 == 0) * unit_length
            return np.array([x, y])
    else:
        raise RuntimeError("There is no built-in cell_place_function for this placement.")

    return cell_place_function


def visualize(circuit: Union[dict, Type[Circuit]],
              placement: Optional[str] = '',
              cell_place_function: Callable = None,
              **kwargs) -> None:
    """Visualize a programmable photonics circuit containing TBUs

    :param circuit: A dictionary contains all TBUs and its connections :param port_magnitude: A np.ndarray of shape (
                    #nodes, len(omega), 2), it will be used if the 'annotate' variable is specified.
    :param placement: a string specified the placement of the provided circuit. It will be used only if
                      cell_place_function is None. :param
    cell_place_function: A callable function, given the cell name, length and width of TBU, return the position of
                        cell center.
    :param kwargs: a set of parameters determining the visualization. Accepted parameters including:

        length_width_ratio: The ratio of TBU length over TBU width
        add_on_ratio: a tuple with two float numbers, the first for length of add-on in the 'TBU width' direction, the
                         second for length of add-on in the 'TBU length' direction.
        annotate: a string represents the annotation style.
        title: the title of the figure.
        line2d_property: a dictionary with properties accepted by plt.plot()
        polygon_property: a dictionary with properties accepted by plt.fill()

    :return: None.

    """

    if placement == '' and cell_place_function is None:
        raise RuntimeError("One of 'notation' and 'cell_place_function' should be provided to do visualization.")

    if isinstance(circuit, Circuit):
        circuit = circuit.circuit_element
    elif not isinstance(circuit, dict):
        raise RuntimeError(
            "The object being visualized must be a dict containing circuit elements or a Circuit instance.")

    cell_place_function = cell_place_function if cell_place_function is not None else _cell_place_function_list(
        placement)
    _visualize_base(circuit, cell_place_function, **kwargs)


def _abc_through_two_points(point1: np.ndarray, point2: np.ndarray) -> np.ndarray:
    """Calculate {a,b,c} of ax+by+c=0 going through two given points.

    :param point1: np.array (2,1) or (2,). Coordinate of point1.
    :param point2: np.array (2,1) or (2,). Coordinate of point2.
    :return: np.array (3,). {a,b,c} in order.
    """

    x1, y1 = point1.flatten()
    x2, y2 = point2.flatten()

    a = y2 - y1
    b = x1 - x2
    c = x2 * y1 - x1 * y2

    ratio = (a ** 2 + b ** 2 + c ** 2) ** 0.5
    a, b, c = a / ratio, b / ratio, c / ratio

    return np.array([a, b, c])


def _perpendicular_one_line_through_one_point(line: np.ndarray, point: np.ndarray) -> np.ndarray:
    """Calculate {a,b,c} of ax+by+c = 0 perpendicular a given line and going through a given point

    :param line: np.array (3,1) or (3,). {a,b,c} of the given line.
    :param point: np.array (2,1) or (2,). Coordinate of the given point.
    :return: np.array (3,). {a,b,c} in order
    """
    line = line.flatten()
    a, b = -line[1], line[0]
    c = - (a * point[0] + b * point[1])

    ratio = (a ** 2 + b ** 2 + c ** 2) ** 0.5
    a, b, c = a / ratio, b / ratio, c / ratio

    return np.array([a, b, c])


def _get_two_edges(tbu_length, tbu_width, cell1_center, cell2_center):
    middle_point = (cell1_center + cell2_center) / 2.0

    direction_vector = cell2_center - cell1_center
    # go along the line connecting cell1_center and cell2_center
    t = tbu_width / 2.0 / np.linalg.norm(direction_vector)
    point1 = middle_point - t * direction_vector
    point2 = middle_point + t * direction_vector

    # go along the direction perpendicular to the line connecting cell1_center and cell2_center
    a, b = -direction_vector[1], direction_vector[0]
    t = tbu_length / 2.0 / (a ** 2 + b ** 2) ** 0.5
    edge1 = np.array([point1 + t * np.array([a, b]), point1 - t * np.array([a, b])])
    edge2 = np.array([point2 + t * np.array([a, b]), point2 - t * np.array([a, b])])

    return edge1, edge2


def _visualize_base(circuit_element: dict, cell_place_func: Callable, **kwargs):
    """"A base function for visualization

    :param circuit_element: A dictionary contains all TBUs and its connections
    :param cell_place_func: A callable function, given the cell name, length and width of TBU, return the position of
                            cell center.
    :param kwargs: a set of parameters determining the visualization. Accepted parameters including:

        length_width_ratio: The ratio of TBU length over TBU width
        add_on_ratio: a tuple with two float numbers, the first for length of add-on in the 'width' direction, the
                    second for length of add-on in the 'length' direction.
        title: the title of the figure.
        line2d_property: a dictionary with properties accepted by plt.plot()
        polygon_property: a dictionary with properties accepted by plt.fill()
        text_property: a dictionary with properties accepted by plt.text()
    :return: None.

    """

    if 'length_width_ratio' not in kwargs.keys():
        length_width_ratio = 0.12
    else:
        length_width_ratio = kwargs['length_width_ratio']

    if 'add_on_ratio' not in kwargs.keys():
        add_on_ratio_width, add_on_ratio_length = 0.4, 0.6
    else:
        add_on_ratio_width, add_on_ratio_length = kwargs['add_on_ratio']

    if 'title' not in kwargs.keys():
        title = ''
    else:
        title = kwargs['title']

    if 'line2d_property' not in kwargs.keys():
        line2d_property = {'linewidth': 3}
    else:
        line2d_property = {**{'linewidth': 3}, **kwargs['line2d_property']}

    if 'polygon_property' not in kwargs.keys():
        polygon_property = {'zorder': 100}
    else:
        polygon_property = {**{'zorder': 100}, **kwargs['polygon_property']}

    if 'text_property' not in kwargs.keys():
        text_property = {'zorder': 100, 'ha': 'center', 'va': 'center'}
    else:
        text_property = {**{'zorder': 100, 'ha': 'center', 'va': 'center'}, **kwargs['text_property']}

    annotate = [] if 'annotate' not in kwargs.keys() else kwargs['annotate']

    tbu_lenth = 1
    tbu_width = tbu_lenth * length_width_ratio

    ax = plt.figure(figsize=[8, 8])
    ax.gca().invert_yaxis()
    plt.axis('equal')
    plt.axis('off')

    for tbu, value in circuit_element.items():
        # as an example, tbu = tbum1_1#2_1#3_v
        cell1_name, cell2_name = tbu.split("_")[1:3]
        cell1_center = cell_place_func(cell1_name, tbu_lenth, tbu_width)
        cell2_center = cell_place_func(cell2_name, tbu_lenth, tbu_width)

        # plot the background two edges, default color is mediumslateblue.
        edge1, edge2 = _get_two_edges(tbu_lenth, tbu_width, cell1_center, cell2_center)
        plt.plot(edge1[:, 0], edge1[:, 1], mcolors.CSS4_COLORS['mediumslateblue'], **line2d_property)
        plt.plot(edge2[:, 0], edge2[:, 1], mcolors.CSS4_COLORS['mediumslateblue'], **line2d_property)

        # plot the add-on rectangle, representing the phase shifts, default color is red.
        add_on_edge1, add_on_edge2 = _get_two_edges(tbu_lenth * (1 - add_on_ratio_length),
                                                    tbu_width * (1 + add_on_ratio_width), cell1_center, cell2_center)
        pts = np.concatenate((add_on_edge1, add_on_edge2), axis=0)
        hull = ConvexHull(pts)
        plt.fill(pts[hull.vertices, 0], pts[hull.vertices, 1], 'red', **polygon_property)

        # deal with different annotation style
        if ('all_ps_symbol_value' in annotate or 'all_ps_value' in annotate) and (
                'theta' not in value.keys() or 'phi' not in value.keys()):
            raise RuntimeError("When 'all_ps_(symbol)_value' is used, phase shift values must be provided.")

        if 'all_ps_symbol_value' in annotate:
            text1_loc, text2_loc = add_on_edge1.mean(axis=0), add_on_edge2.mean(axis=0)
            theta_cell_name, phi_cell_name = value['ln'][0].split('_')[1], value['ln'][1].split('_')[1]
            plt.text(*text1_loc,
                     r'$\theta=%.1f\pi$' % (
                             value['theta'] / np.pi) if theta_cell_name == cell1_name else r'$\phi = %.1f\pi$' % (
                             value['phi'] / np.pi),
                     **text_property)
            plt.text(*text2_loc,
                     r'$\theta=%.1f\pi$' % (
                             value['theta'] / np.pi) if theta_cell_name == cell2_name else r'$\phi=%.1f\pi$' % (
                             value['phi'] / np.pi),
                     **text_property)

        if 'all_ps_symbol' in annotate:
            text1_loc, text2_loc = add_on_edge1.mean(axis=0), add_on_edge2.mean(axis=0)
            theta_cell_name, phi_cell_name = value['ln'][0].split('_')[1], value['ln'][1].split('_')[1]
            plt.text(*text1_loc,
                     r'$\theta$' if theta_cell_name == cell1_name else r'$\phi$',
                     **text_property)
            plt.text(*text2_loc,
                     r'$\theta$' if theta_cell_name == cell2_name else r'$\phi$',
                     **text_property)

        if 'all_ps_value' in annotate:
            text1_loc, text2_loc = add_on_edge1.mean(axis=0), add_on_edge2.mean(axis=0)
            theta_cell_name, phi_cell_name = value['ln'][0].split('_')[1], value['ln'][1].split('_')[1]
            plt.text(*text1_loc,
                     r'$%.1f\pi$' % (value['theta'] / np.pi) if theta_cell_name == cell1_name else r'$%.1f\pi$' % (
                             value['phi'] / np.pi),
                     **text_property)
            plt.text(*text2_loc,
                     r'$%.1f\pi$' % (value['theta'] / np.pi) if theta_cell_name == cell2_name else r'$%.1f\pi$' % (
                             value['phi'] / np.pi),
                     **text_property)

        if 'all_cell' in annotate:
            plt.text(*cell1_center, '(' + cell1_name.replace('#', ',') + ')', **text_property)
            plt.text(*cell2_center, '(' + cell2_name.replace('#', ',') + ')', **text_property)

        if 'all_tbu_name' in annotate:
            plt.text(*(cell1_center + cell2_center) / 2.0, tbu, **text_property)

    if title:
        plt.title(title)

    plt.show()


if __name__ == '__main__':
    circuit_element = generate('square_1', [3, 2], init_dict={'theta': 0, 'phi': 0.5 * np.pi})
    visualize(circuit_element, 'square_1', line2d_property={}, polygon_property={},
              annotate=['all_cell', 'all_ps_symbol'])

    circuit_element = generate('hexagonal_1', [2], init_dict={'theta': 0, 'phi': 0.5 * np.pi})
    visualize(circuit_element, 'hexagonal_1', line2d_property={}, polygon_property={},
              annotate=['all_cell', 'all_ps_symbol'])

    circuit_element = generate('triangular_1', [3, 5], init_dict={'theta': 0, 'phi': 0.5 * np.pi})
    visualize(circuit_element, 'triangular_1', line2d_property={}, polygon_property={},
              annotate=['all_cell', 'all_ps_symbol'], text_property={})



import numpy as np
import matplotlib.pyplot as plt
from spode.core.solver import Circuit
from typing import Optional, Union, List, Set, Tuple, Any, Callable, Type
from spode.util.generator import _config_size_dict, Generator
from scipy.spatial import ConvexHull
import matplotlib.colors as mcolors

__all__ = ['visualize']

def _cell_place_function_list(placement: str):
    if placement.lower() == 'square_1':
        def cell_place_function(x, tbu_length, tbu_width):
            return np.array([tbu_width * (x[0] + 1) + tbu_length * (x[0] + 0.5),
                             tbu_width * (x[1] + 1) + tbu_length * (x[1] + 0.5)])

    return cell_place_function


def visualize(circuit: Union[dict, Type[Circuit]],
              notation: Optional[str] = '',
              cell_place_function: Callable = None,
              **kwargs):
    if isinstance(circuit, dict):
        if cell_place_function is not None:
            _visualize_base(circuit, cell_place_function, **kwargs)
        else:
            cell_place_function = _cell_place_function_list(notation)
            _visualize_base(circuit, cell_place_function, **kwargs)
    elif isinstance(circuit, Circuit):
        print("hello world")
    else:
        raise RuntimeError(
            "The object being visualized must be a dict containing circuit elements or a Circuit instance.")


def _abc_through_two_points(point1: np.ndarray, point2: np.ndarray) -> np.ndarray:
    """Calculate {a,b,c} of ax+by+c=0 going through two given points.

    :param point1: np.array (2,1) or (2,). Coordinate of point1.
    :param point2: np.array (2,1) or (2,). Coordinate of point2.
    :return: np.array (3,). {a,b,c} in order.
    """

    x1, y1 = point1.flatten()
    x2, y2 = point2.flatten()

    a = y2 - y1
    b = x1 - x2
    c = x2 * y1 - x1 * y2

    ratio = (a ** 2 + b ** 2 + c ** 2) ** 0.5
    a, b, c = a / ratio, b / ratio, c / ratio

    return np.array([a, b, c])


def _perpendicular_one_line_through_one_point(line: np.ndarray, point: np.ndarray) -> np.ndarray:
    """Calculate {a,b,c} of ax+by+c = 0 perpendicular a given line and going through a given point

    :param line: np.array (3,1) or (3,). {a,b,c} of the given line.
    :param point: np.array (2,1) or (2,). Coordinate of the given point.
    :return: np.array (3,). {a,b,c} in order
    """
    line = line.flatten()
    a, b = -line[1], line[0]
    c = - (a * point[0] + b * point[1])

    ratio = (a ** 2 + b ** 2 + c ** 2) ** 0.5
    a, b, c = a / ratio, b / ratio, c / ratio

    return np.array([a, b, c])


def _get_two_edges(tbu_length, tbu_width, cell1_center, cell2_center):
    middle_point = (cell1_center + cell2_center) / 2.0
    a, b, c = _abc_through_two_points(cell1_center, cell2_center)

    # go along the line connecting cell1_center and cell2_center
    t = tbu_width / 2.0 / (a ** 2 + b ** 2) ** 0.5
    point1 = middle_point - t * np.array([b, -a])
    point2 = middle_point + t * np.array([b, -a])

    # go along the direction perpendicular to the line connecting cell1_center and cell2_center
    t = tbu_length / 2 / (a ** 2 + b ** 2) ** 0.5
    edge1 = np.array([point1 + t * np.array([a, b]), point1 - t * np.array([a, b])])
    edge2 = np.array([point2 + t * np.array([a, b]), point2 - t * np.array([a, b])])

    return edge1, edge2


def _visualize_base(circuit_element: dict,
                    cell_place_func: Callable,
                    length_width_ratio: Optional[float] = 0.12,
                    add_on_ratio: Optional[dict] = (0.4, 0.6),
                    title: Optional[str] = '',
                    line2d_property: Optional[dict] = None,
                    polygon_property: Optional[dict] = None):
    """"A base function for visualization

    :param circuit_element: A dictionary contains all TBUs and its connections
    :param length_width_ratio: The ratio of TBU length over TBU width
    :param add_on_ratio: a tuple with two float numbers, the first for length of add-on in the 'width' direction, the
                         second for length of add-on in the 'length' direction.
    :param title: the title of the figure.
    :param line2d_property: a dictionary contains all properties accepted by plt.plot()
    :param polygon_property: a dictionary contains all properties accepted by plt.fill()

    :return: None.

    """
    tbu_lenth = 1
    tbu_width = tbu_lenth * length_width_ratio
    add_on_ratio_width, add_on_ratio_length = add_on_ratio

    ax = plt.figure(figsize=[8, 8])
    ax.gca().invert_yaxis()
    plt.axis('equal')
    plt.axis('off')

    for tbu, value in circuit_element.items():
        # as an example, tbu = tbum1_1#2_1#3_v
        tbu_name, cell1_name, cell2_name, *_ = tbu.split("_")
        cell1_center = cell_place_func([int(ele) for ele in cell1_name.split('#')])
        cell2_center = cell_place_func([int(ele) for ele in cell2_name.split('#')])

        # plot the background two edges, default color is mediumslateblue.
        edge1, edge2 = _get_two_edges(tbu_lenth, tbu_width, cell1_center, cell2_center)
        plt.plot(edge1[:, 0], edge1[:, 1], mcolors.CSS4_COLORS['mediumslateblue'], **line2d_property)
        plt.plot(edge2[:, 0], edge2[:, 1], mcolors.CSS4_COLORS['mediumslateblue'], **line2d_property)

        # plot the add-on rectangle, representing the phase shifts, default color is red.
        add_on_edge1, add_on_edge2 = _get_two_edges(tbu_lenth * (1 - add_on_ratio_length),
                                                    tbu_width * (1 + add_on_ratio_width), cell1_center, cell2_center)
        pts = np.concatenate((add_on_edge1, add_on_edge2), axis=0)
        hull = ConvexHull(pts)
        plt.fill(pts[hull.vertices, 0], pts[hull.vertices, 1], 'red', **{**{'zorder': 100}, **polygon_property})

    if title:
        plt.title(title)

    plt.show()


# def _visualize_square_1(circuit_element: dict, cell_place_func: Callable, length_width_ratio: Optional[float] = 0.12,
#                         add_on_ratio: Optional[dict] = (0.4, 0.6), title: Optional[str] = '',
#                         line2d_property: Optional[dict] = None, polygon_property: Optional[dict] = None):
#     _visualize_base()


if __name__ == '__main__':
    # circuit_element = {
    #     'tbum1': {'Ln': ['N0', 'n1'], 'rn': ['n2', 'N3'], 'theta': 0.0 * np.pi, 'pHi': 0.0 * np.pi, 'L': 250e-6},
    #     'wg1': {'ln': ['n3'], 'rn': ['n4'], 'L': 0, 'alpha': 1.0},
    #     'pS1': {'ln': ['n4'], 'rn': ['n5'], 'ps': 0.0 * np.pi}}
    #
    # mode_info = {'neFf': 2.35, 'wl': 1550e-9}
    # omega = np.linspace(192.5, 193.5, 3) * 1e12 * 2 * np.pi
    # prob_node = ['N5']
    # deri_node = ['n5']
    # srce_node = {'N0': 1.0}
    # deri_vari = ['tbum1::theta', 'wG1::l']
    #
    # circ_instance = Circuit(circuit_element=circuit_element,
    #                         mode_info=mode_info,
    #                         omega=omega,
    #                         srce_node=srce_node,
    #                         prob_node=prob_node,
    #                         deri_node=deri_node,
    #                         deri_vari=deri_vari)
    # visualize(circ_instance)

    generator = Generator()
    circuit_element = generator.generate('square_1', [5, 5], init_dict={})
    _visualize_square_1(circuit_element, line2d_property={'linewidth': 3}, polygon_property={})




import numpy as np
from typing import Optional, Union, List, Set, Tuple, Any, Callable, Type
import importlib
import json
from functools import reduce
import operator

from spode.core.model import _model_json_name

__all__ = ['Circuit']


# TODO: currently only support constant source, is frequency-dependent source meaningful?

class Circuit(object):
    """The Circuit class object"""

    def __init__(self, circuit_element: dict, mode_info: dict, omega: Union[List[float], np.ndarray],
                 srce_node: dict, prob_node: List[str],
                 deri_node: Optional = None, deri_vari: Optional = None, notation: Optional[str] = ''):
        """Initialize the Circuit class.

        The Circuit object takes a few parameters as input, and then initialize its attributes. In the meanwhile, we will
        also do some preprocessing in the initialization process. The following parameters will be directly stored in the
        Circuit object using the same name.

        :param circuit_element: information about elements and their connections in the concerned circuit.
        :param mode_info: information about the simulating propagating mode, e.g., {'neff':2.0}.
        :param omega: the simulation angular frequency grid points.
        :param srce_node: the node where the sources are injected and the values of sources.
        :param probe_node: specify which ports' complex magnitudes are of interests.
        :param deri_node: specify which port magnitude derivatives are of interests.


        These attributes will be calculated based on the above given parameters. Moreover, in the calculation, we will do some checking about the validity of the input parameters, in case of erroneous usage. Note that our checking might not be comprehensive.

        info: information about all existing models (built-in and user-defined ones).
        deri_vari_dict: specify which attribute of which circuit element's will be calculated derivative w.r.t.
        node2ind, ind2node: information about the mapping from node's string name and node's index, and vice versa.
        node_element: specify which circuit elements are conneted to a node.
        inward_node, outward_node: information about the inward and outward direction of the port of a circuit element.

        """

        # eval(repr().lower()) will convert every string to the corresponding lower case.
        try:
            self.circuit_element = eval(repr(circuit_element).lower())
            self.mode_info = eval(repr(mode_info).lower())
            self.omega = omega
            self.srce_node = eval(repr(srce_node).lower())
            self.prob_node = eval(repr(prob_node).lower())
            self.deri_node = eval(repr(deri_node).lower())
            self.deri_vari = eval(repr(deri_vari).lower())
            self.notation = eval(repr(notation).lower())
        except:
            raise RuntimeError("The provided inputs to initialize Circuit instance are problematic.")

        # The Circuit instance needs re-initialization after calling register_model(), so that self.info could be updated.
        with open(_model_json_name, 'r') as f:
            self.info = json.load(f)

        self._preprocess()

    def _preprocess(self):
        """Preprocess at the end of initialization

        We need to preprocess at the end of initialization to do some checking, and generate some useful variables. For
        details, please refer to each of the used functions. Moreover, this _preprocess function needs to be called after
        calling update_attr().

        """
        self.deri_vari_dict = self._deri_vari_processing()
        self.node2ind, self.ind2node, self.node_element, self.inward_node, self.outward_node = self._node_processing()

    def _check_name(self, index_string: str, attr_name: str, criterion: Optional[str] = 'identical') -> bool:
        """Check whether a given string occurs in an attribute of self.info

        This helper function is used to check the provided string by the user is correct. For instance, in the model.json
        file, we have a circuit element with model_name=wg, so that we have:

        _check_name('wgxxx', 'model_name', 'startswith') return True
        _check_name('wgxxx', 'model_name', 'identical') return False
        _check_name('wg', 'model_name', 'startswith') return True
        _check_name('wg', 'model_name', 'identical') return True

        :param index_string: a string that will be examined if it exists in self.info.
        :param attr_name: specify which attributes are of interests. It has to be those attributes in model.json, such as
                          'model_name', 'class_name', 'required_attr', 'optional_attr'.
        :param criterion: specify the comparison criterion.
        :return: True for occurrence, false for not.
        """
        if criterion not in ['identical', 'startswith']:
            raise RuntimeError("The comparison criterion is not implemented.")

        for entry in self.info:
            base_string = entry[attr_name]
            if criterion == 'startswith' and index_string.startswith(base_string):
                return True
            if criterion == 'identical' and base_string == index_string:
                return True

        return False

    def _deri_vari_processing(self) -> dict:
        """Preprocess the derivative information

        Process the attributes of Circuit class, and return a dictionary about which circuit element's attribute will be
         calculated derivative w.r.t. The dictionary key represents the circuit element, and the dictionary value
         represents the attribute. A few examinations are performed in this function to eliminate the possibility of
         erroneous usage of the simulator, but the readers should be warned that our checking is not comprehensive.

        :return: A dictionary with circuit element as key, and its attribute as value.
        """
        deri_dict = {}
        for cur_deri_vari in self.deri_vari:
            try:
                key, value = cur_deri_vari.split("::")
            except:
                raise RuntimeError("The provided derivative variable '%s' is problematic." % cur_deri_vari)

            if self._check_name(key, 'model_name', 'startswith'):
                if key not in self.circuit_element.keys():
                    raise RuntimeError(
                        "The element named '%s' (required by '%s') doesn't present in the circuit (i.e., circuit_element)." % (
                        key, cur_deri_vari))
                elif key not in deri_dict.keys():
                    # We postpone the checking of whether 'value' exists to when we calculate the derivative.
                    deri_dict[key] = [value]
                elif value in deri_dict[key]:
                    raise RuntimeError("The derivative variable '%s' is replicated." % cur_deri_vari)
                else:
                    # We postpone the checking of whether 'value' exists to when we calculate the derivative.
                    deri_dict[key].append(value)
            else:
                raise RuntimeError("The circuit model of '%s' (required by '%s') is not defined. Please check if it "
                                   "is a typo." % (key, cur_deri_vari))

        return deri_dict

    def _node_processing(self) -> Tuple[dict, dict, dict, dict, dict]:
        """Preprocess the node information

        Process the attributes of Circuit class, and return several dictionaries about the mappings of nodes. The nodes
        are specified in strings in the netlist, while to run the simulation, we need to assign each node an index, so
         that a matrix equation could be built. We need to store the forward mapping from node string name to node index
          (node2ind), as well as the backward mapping from node index to node string name (ind2node). Moreover, we also
          need to store information such as which circuit elements are connected to a node (node_element), and the inward
         and outward optical signals occupying which index (inward_node, outward_node). Except inward_node and outward_node,
         the other dictionaries should be easily understood. For these two dictionaries, please refer to the document on
         Principle of Simulator (PoS).

        :return: a tuple of dictionaries.
        """
        node_index, node2ind, ind2node, node_has_ele = 0, {}, {}, {}
        for ele, attr in self.circuit_element.items():
            nodes = attr['ln'] + attr['rn']
            for node in nodes:
                # update node2ind, ind2node
                if node not in node2ind:
                    node2ind[node] = node_index
                    ind2node[node_index] = node
                    node_index += 1
                # update node_has_ele
                if node not in node_has_ele:
                    node_has_ele[node] = [ele]
                else:
                    node_has_ele[node].append(ele)

        # build the inward node index dict and outward node index dict
        inward_node, outward_node = {}, {}
        for ele, attr in self.circuit_element.items():
            inward_node[ele], outward_node[ele] = {'ln': [], 'rn': []}, {'ln': [], 'rn': []}

            # ele_left_nodes, ele_right_nodes = attr['ln'], attr['rn']

            pair = zip(['ln', 'rn'], [attr['ln'], attr['rn']])

            for string, ele_nodes in pair:
                for node in ele_nodes:
                    if node_has_ele[node][0] == ele:
                        inward_node[ele][string].append(node2ind[node] * 2)
                        outward_node[ele][string].append(node2ind[node] * 2 + 1)
                    elif node_has_ele[node][1] == ele:
                        inward_node[ele][string].append(node2ind[node] * 2 + 1)
                        outward_node[ele][string].append(node2ind[node] * 2)
                    else:
                        raise RuntimeError("The node '%s' has more than two circuit elements connected." % (node))

            # for node in ele_left_nodes:
            #     if node_has_ele[node][0] == ele:
            #         inward_node[ele]['ln'].append(node2ind[node] * 2)
            #         outward_node[ele]['ln'].append(node2ind[node] * 2 + 1)
            #     elif node_has_ele[node][1] == ele:
            #         inward_node[ele]['ln'].append(node2ind[node] * 2 + 1)
            #         outward_node[ele]['ln'].append(node2ind[node] * 2)
            #     else:
            #         raise RuntimeError("The node '%s' has more than two circuit elements connected." % (node))
            # for node in ele_right_nodes:
            #     if node_has_ele[node][0] == ele:
            #         inward_node[ele]['rn'].append(node2ind[node] * 2)
            #         outward_node[ele]['rn'].append(node2ind[node] * 2 + 1)
            #     elif node_has_ele[node][1] == ele:
            #         inward_node[ele]['rn'].append(node2ind[node] * 2 + 1)
            #         outward_node[ele]['rn'].append(node2ind[node] * 2)
            #     else:
            #         raise RuntimeError("The node '%s' has more than two circuit elements connected." % (node))

        # As we have now processed all nodes, we could do a check if self.prob_nodes are correctly provided.

        if not set(self.srce_node.keys()).issubset(set(node2ind.keys())):
            raise RuntimeError("'srce_node' contains node not included in the circuit (i.e., circuit_element).")
        if not set(self.deri_node).issubset(set(node2ind.keys())):
            raise RuntimeError("'deri_node' contains node not included in the circuit (i.e., circuit_element).")
        if not set(self.prob_node).issubset(set(node2ind.keys())):
            raise RuntimeError("'prob_node' contains node not included in the circuit (i.e., circuit_element).")

        return node2ind, ind2node, node_has_ele, inward_node, outward_node

    def _build_matrix(self, A: np.ndarray, inward_node: List[int], outward_node: List[int], smatrix: np.ndarray,
                      line_counter: int, fillconst: float) -> Tuple[np.ndarray, int]:
        """Embed the relation (constraint) of one S matrix into the overall system matrix A.

        This is a helper function used by the solve method. A and smatrix is a three-dimensional np.ndarray. Both their
        shape[0] equals to the length of omega. Namely, A is a batched system matrix. Please refer to our document on
        Principle of Simulator (PoS) for details.

        :param A: the system matrix which the S matrix will be embedded into.
        :param inward_node: the node indices for the inward direction.
        :param outward_node: the node indices for the outward direction.
        :param smarix: the current S matrix.
        :param line_counter: specify which row the relation (constraint) will be at.
        :param fillconst: specify the filled value associated with the output node. It will be set to 1 when calculating
                          node magnitude response, and 0 when calculating derivative.
        :return: the updated system matrix A and the line counter.
        """
        for i in range(len(outward_node)):
            A[:, line_counter, outward_node[i]] = fillconst
            A[:, line_counter, inward_node] = smatrix[:, i, :]
            line_counter += 1
        return A, line_counter

    def solve(self, require_grads: Optional = True) -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:
        """Solve the circuit port magnitude and derivative.

        This function is the core of the Circuit class. The solver logic is to build a large system matrix Ax=b at one
        frequency point, with A and b embedding all S-matrix relations of all circuit elements, and x is a (2N,1) unknown
        vector represents the complex port magnitudes. Here N represents the number of ports, because for each port, the
         optical signals have two propagating directions. Derivative is calculated based on a first order Taylor
        expansion: delta_A * x + A * delta_x = 0 => delta_x = -inv(A)* delta_A * x.

        In implementation, the port magnitudes and derivatives are calculated in a batch manner: all angular frequency
        points are solved at the same time. Namely, A is of size (len(omega), 2N, 2N), etc.

        :param require_grads: Whether gradient is calculated.
        :return: The port magnitude, and the gradient if needed.
        """
        require_grads = require_grads and self.deri_vari is not None and self.deri_node is not None \
                        and len(self.deri_vari) >= 1 and len(self.deri_node) >= 1

        line_counter, num_node = 0, len(self.ind2node.keys())

        module = importlib.import_module("spode.core.model")

        A, b = np.zeros([len(self.omega), 2 * num_node, 2 * num_node], dtype=np.complex_), \
               np.zeros([len(self.omega), 2 * num_node, 1], dtype=np.complex_)

        for node, ele in self.node_element.items():
            if len(ele) == 1:
                src_value = 0.0 if node not in self.srce_node.keys() else self.srce_node[node]

                # source must be injected at floating node, and the inward direction takes the first place
                A[:, line_counter, 2 * self.node2ind[node]] = 1.0
                b[:, line_counter, 0] = src_value
                line_counter += 1

        grads = [np.zeros([len(self.omega), 2 * num_node, 2 * num_node], dtype=np.complex_)
                 for _ in range(len(self.deri_vari))]

        for ele, attr in self.circuit_element.items():
            for entry in self.info:
                if ele.startswith(entry['model_name'].lower()):
                    class_ = getattr(module, entry['class_name'])
                    ele_instance = class_(**{**attr, **self.mode_info})
                    if ele in self.deri_vari_dict.keys():

                        S_l2r, grad_l2r_dict = ele_instance.get_smatrix(self.omega, direction='l2r',
                                                                        deri_vari=self.deri_vari_dict[ele])
                        S_r2l, grad_r2l_dict = ele_instance.get_smatrix(self.omega, direction='r2l',
                                                                        deri_vari=self.deri_vari_dict[ele])

                        for cur_deri_vari in self.deri_vari_dict[ele]:
                            if cur_deri_vari not in grad_r2l_dict.keys() or cur_deri_vari not in grad_l2r_dict.keys():
                                raise RuntimeError(
                                    "The gradient of S matrix w.r.t. '%s' in model '%s' is not implemented"
                                    % (cur_deri_vari, entry['class_name']))

                            line_counter_wrk = line_counter
                            index = self.deri_vari.index(ele + "::" + cur_deri_vari)
                            grads[index], line_counter_wrk = self._build_matrix(grads[index],
                                                                                self.inward_node[ele]['ln'],
                                                                                self.outward_node[ele]['rn'],
                                                                                grad_l2r_dict[cur_deri_vari],
                                                                                line_counter_wrk, 0)
                            grads[index], line_counter_wrk = self._build_matrix(grads[index],
                                                                                self.inward_node[ele]['rn'],
                                                                                self.outward_node[ele]['ln'],
                                                                                grad_r2l_dict[cur_deri_vari],
                                                                                line_counter_wrk, 0)

                    else:
                        S_l2r = ele_instance.get_smatrix(self.omega, direction='l2r', deri_vari=None)
                        S_r2l = ele_instance.get_smatrix(self.omega, direction='r2l', deri_vari=None)

                    A, line_counter = self._build_matrix(A, self.inward_node[ele]['ln'], self.outward_node[ele]['rn'],
                                                         S_l2r,
                                                         line_counter, -1.0)
                    A, line_counter = self._build_matrix(A, self.inward_node[ele]['rn'], self.outward_node[ele]['ln'],
                                                         S_r2l,
                                                         line_counter, -1.0)
                    break
            else:
                raise RuntimeError("The model %s is not defined in the simulator. Check if there is a typo, "
                                   "or define the model by yourself." % (ele))

        try:
            invA = np.linalg.inv(A)
        except:
            raise RuntimeError("Solving the system matrix fails. Please check the provided netlist.")

        response = np.matmul(invA, b)

        returned_res = np.zeros([len(self.prob_node), len(self.omega), 2], dtype=np.complex_)
        for i in range(len(self.prob_node)):
            # usually the probed node is a floating node, so the first place contains the complex magnitude for the
            # inward direction; the second place contains that for the outward direction.
            node_ind = self.node2ind[self.prob_node[i]]
            returned_res[i, :, 0] = response[:, 2 * node_ind, 0]
            returned_res[i, :, 1] = response[:, 2 * node_ind + 1, 0]

        if not require_grads:
            return returned_res

        returned_grads = np.zeros([len(self.deri_node), len(self.deri_vari), len(self.omega), 2], dtype=np.complex_)
        for j in range(len(self.deri_vari)):
            wrk_grad = np.matmul(np.matmul(-invA, grads[j]), response)
            for i in range(len(self.deri_node)):
                ind = self.node2ind[self.deri_node[i]]
                returned_grads[i, j, :, :] = wrk_grad[:, 2 * ind:2 * ind + 2, 0]

        return returned_res, returned_grads

    def update_attr(self, update_attr: str, new_value: Any, key_strings: Optional[str] = None):
        """Update the attributes of the circuit

        Update the specified attribute with a provided new value. Since the Circuit class has both dict (even nested dict)
        and list attributes, this function provides a convenient interface to update any attribute. Example usage:

        circuit_instance.update_attr('mode_info', 4, "neff")
        circuit_instance.update_attr('deri_node', 'n222')
        circuit_instance.update_attr('circuit_element', ['n11', 'n222'], key_strings='tbum1::ln')
        circuit_instance.update_attr('circuit_element', 0.3 * np.pi, key_strings='ps1::ps')

        :param update_attr: the attribute that will be updated.
        :param new_value: the new value of the attribute.
        :param key_strings: It specify the key, if the updated attribute is a dict.
        """
        update_attr = update_attr.lower()

        cur_value = getattr(self, update_attr)
        if isinstance(cur_value, dict):
            key_strings = key_strings.lower().split("::")
            reduce(operator.getitem, key_strings[:-1], cur_value)[key_strings[-1]] = new_value
            setattr(self, update_attr, cur_value)
        else:
            setattr(self, update_attr, new_value)

        # if the node or derivative variable is updated, we need to call _preprocess().
        self._preprocess()

    def get_attr(self, get_attr: str, key_strings: Optional[str] = None) -> Any:
        """Get the current value of some attribute.

        Get the current value of some attribute specified by get_attr (as well as key_strings if it exists). Example usage:

        circuit_instance.get_attr('circuit_element', 'tbum1::alpha')
        circuit_instance.get_attr('mode_info')

        :param get_attr: specify the value of which attribute we will get.
        :param: key_strings: If the desired attribute is the value in a dict, the key needs to be provided.
        :return: the value of the attribute.
        """
        get_attr = get_attr.lower()

        cur_value = getattr(self, get_attr)
        if isinstance(cur_value, dict):
            key_strings = key_strings.lower().split("::")
            return reduce(operator.getitem, key_strings, cur_value)
        return cur_value


if __name__ == '__main__':
    circuit_element = {
        'tbum1': {'Ln': ['N0', 'n1'], 'rn': ['n2', 'N3'], 'theta': 0.0 * np.pi, 'pHi': 0.0 * np.pi, 'L': 250e-6},
        'wg1': {'ln': ['n3'], 'rn': ['n4'], 'L': 0, 'alpha': 1.0},
        'pS1': {'ln': ['n4'], 'rn': ['n5'], 'ps': 0.0 * np.pi}}

    mode_info = {'neFf': 2.35, 'wl': 1550e-9}
    omega = np.linspace(192.5, 193.5, 3) * 1e12 * 2 * np.pi
    prob_node = ['N5']
    deri_node = ['n5']
    srce_node = {'N0': 1.0}
    deri_vari = ['tbum1::theta', 'wG1::l']

    circ_instance = Circuit(circuit_element=circuit_element,
                            mode_info=mode_info,
                            omega=omega,
                            srce_node=srce_node,
                            prob_node=prob_node,
                            deri_node=deri_node,
                            deri_vari=deri_vari)

    # print(circ_instance.node2ind)
    # print(circ_instance.ind2node)
    # print(circ_instance.node_element)
    # print(circ_instance.inward_node)
    # print(circ_instance.outward_node)

    returned = circ_instance.solve()

    print("hello")

    # circ_instance.update_attr('mode_info', 4, "neff")
    # circ_instance.update_attr('deri_node', 'n222')
    # circ_instance.update_attr('circuit_element', ['n11', 'n222'], key_strings='tbum1::ln')
    # circ_instance.update_attr('circuit_element', 0.3 * np.pi, key_strings='ps1::ps')
    #
    # circ_instance.get_attr('mode_info', "neff")
    # circ_instance.get_attr('mode_info', "neff")
    # circ_instance.get_attr('circuit_element', key_strings='tbum1::ln')
    # circ_instance.get_attr('circuit_element', key_strings='ps1::ps')



from setuptools import setup, find_packages

setup(
        name='spode',
        version='0.0.7',
        description='A simulator with programmable photonics and differentiability emphasis',
        author='ZhengqiGao',
        author_email='zhenqi@mit.edu',
        url='https://github.com/zhengqigao/spode',
        long_description=open('README.md').read(),
        long_description_content_type='text/markdown',
        install_requires = ['numpy'],
        license = 'GPL-3.0',
        package_data ={'spode': ['core/model.json']},
        python_requires='>=3',
        packages=find_packages()
)

